@@ -1,8 +1,10 @@
+"""nbclient implementation."""
 import asyncio
 import atexit
 import base64
 import collections
 import datetime
+import re
 import signal
 import typing as t
 from contextlib import asynccontextmanager, contextmanager
@@ -14,18 +16,7 @@ from jupyter_client import KernelManager
 from jupyter_client.client import KernelClient
 from nbformat import NotebookNode
 from nbformat.v4 import output_from_msg
-from traitlets import (
-    Any,
-    Bool,
-    Callable,
-    Dict,
-    Enum,
-    Integer,
-    List,
-    Type,
-    Unicode,
-    default,
-)
+from traitlets import Any, Bool, Callable, Dict, Enum, Integer, List, Type, Unicode, default
 from traitlets.config.configurable import LoggingConfigurable
 
 from .exceptions import (
@@ -38,8 +29,12 @@ from .exceptions import (
 from .output_widget import OutputWidget
 from .util import ensure_async, run_hook, run_sync
 
+_RGX_CARRIAGERETURN = re.compile(r".*\r(?=[^\n])")
+_RGX_BACKSPACE = re.compile(r"[^\n]\b")
+
 
 def timestamp(msg: t.Optional[t.Dict] = None) -> str:
+    """Get the timestamp for a message."""
     if msg and 'header' in msg:  # The test mocks don't provide a header, so tolerate that
         msg_header = msg['header']
         if 'date' in msg_header and isinstance(msg_header['date'], datetime.datetime):
@@ -52,7 +47,7 @@ def timestamp(msg: t.Optional[t.Dict] = None) -> str:
                     formatted_time
                 ):  # docs indicate strftime may return empty string, so let's catch that too
                     return formatted_time
-            except Exception:
+            except Exception:  # noqa
                 pass  # fallback to a local time
 
     return datetime.datetime.utcnow().isoformat() + 'Z'
@@ -435,6 +430,14 @@ class NotebookClient(LoggingConfigurable):
         )
     )
 
+    coalesce_streams = Bool(
+        help=dedent(
+            """
+            Merge all stream outputs with shared names into single streams.
+            """
+        )
+    )
+
     def __init__(self, nb: NotebookNode, km: t.Optional[KernelManager] = None, **kw: t.Any) -> None:
         """Initializes the execution manager.
 
@@ -556,13 +559,11 @@ class NotebookClient(LoggingConfigurable):
         try:
             self.kc = self.km.client()
             await ensure_async(self.kc.start_channels())  # type:ignore[func-returns-value]
-            await ensure_async(
-                self.kc.wait_for_ready(timeout=self.startup_timeout)  # type:ignore[attr-defined]
-            )
+            await ensure_async(self.kc.wait_for_ready(timeout=self.startup_timeout))
         except Exception as e:
             self.log.error(
                 "Error occurred while starting new kernel client for kernel {}: {}".format(
-                    self.km.kernel_id, str(e)
+                    getattr(self.km, 'kernel_id', None), str(e)
                 )
             )
             await self._async_cleanup_kernel()
@@ -626,7 +627,8 @@ class NotebookClient(LoggingConfigurable):
         atexit.register(self._cleanup_kernel)
 
         def on_signal():
-            asyncio.ensure_future(self._async_cleanup_kernel())
+            """Handle signals."""
+            self._async_cleanup_kernel_future = asyncio.ensure_future(self._async_cleanup_kernel())
             atexit.unregister(self._cleanup_kernel)
 
         loop = asyncio.get_event_loop()
@@ -710,6 +712,7 @@ class NotebookClient(LoggingConfigurable):
     execute = run_sync(async_execute)
 
     def set_widgets_metadata(self) -> None:
+        """Set with widget metadata."""
         if self.widget_state:
             self.nb.metadata.widgets = {
                 'application/vnd.jupyter.widget-state+json': {
@@ -759,7 +762,6 @@ class NotebookClient(LoggingConfigurable):
         task_poll_output_msg: asyncio.Future,
         task_poll_kernel_alive: asyncio.Future,
     ) -> t.Dict:
-
         msg: t.Dict
         assert self.kc is not None
         new_timeout: t.Optional[float] = None
@@ -784,7 +786,7 @@ class NotebookClient(LoggingConfigurable):
                             task_poll_kernel_alive.cancel()
                             raise CellTimeoutError.error_from_timeout_and_cell(
                                 "Timeout waiting for IOPub output", self.iopub_timeout, cell
-                            )
+                            ) from None
                         else:
                             self.log.warning("Timeout waiting for IOPub output")
                     task_poll_kernel_alive.cancel()
@@ -802,7 +804,6 @@ class NotebookClient(LoggingConfigurable):
     async def _async_poll_output_msg(
         self, parent_msg_id: str, cell: NotebookNode, cell_index: int
     ) -> None:
-
         assert self.kc is not None
         while True:
             msg = await ensure_async(self.kc.iopub_channel.get_msg(timeout=None))
@@ -837,7 +838,6 @@ class NotebookClient(LoggingConfigurable):
     async def _async_handle_timeout(
         self, timeout: int, cell: t.Optional[NotebookNode] = None
     ) -> t.Union[None, t.Dict]:
-
         self.log.error("Timeout waiting for execute reply (%is)." % timeout)
         if self.interrupt_on_timeout:
             self.log.error("Interrupting kernel")
@@ -862,7 +862,7 @@ class NotebookClient(LoggingConfigurable):
     async def async_wait_for_reply(
         self, msg_id: str, cell: t.Optional[NotebookNode] = None
     ) -> t.Optional[t.Dict]:
-
+        """Wait for a message reply."""
         assert self.kc is not None
         # wait for finish, with timeout
         timeout = self._get_timeout(cell)
@@ -895,7 +895,6 @@ class NotebookClient(LoggingConfigurable):
     async def _check_raise_for_error(
         self, cell: NotebookNode, cell_index: int, exec_reply: t.Optional[t.Dict]
     ) -> None:
-
         if exec_reply is None:
             return None
 
@@ -1003,7 +1002,7 @@ class NotebookClient(LoggingConfigurable):
         except asyncio.CancelledError:
             # can only be cancelled by task_poll_kernel_alive when the kernel is dead
             task_poll_output_msg.cancel()
-            raise DeadKernelError("Kernel died")
+            raise DeadKernelError("Kernel died") from None
         except Exception as e:
             # Best effort to cancel request if it hasn't been resolved
             try:
@@ -1019,6 +1018,44 @@ class NotebookClient(LoggingConfigurable):
             self.on_cell_executed, cell=cell, cell_index=cell_index, execute_reply=exec_reply
         )
         await self._check_raise_for_error(cell, cell_index, exec_reply)
+
+        if self.coalesce_streams and cell.outputs:
+            new_outputs = []
+            streams: dict[str, NotebookNode] = {}
+            for output in cell.outputs:
+                if output["output_type"] == "stream":
+                    if output["name"] in streams:
+                        streams[output["name"]]["text"] += output["text"]
+                    else:
+                        new_outputs.append(output)
+                        streams[output["name"]] = output
+                else:
+                    new_outputs.append(output)
+
+            # process \r and \b characters
+            for output in streams.values():
+                old = output["text"]
+                while len(output["text"]) < len(old):
+                    old = output["text"]
+                    # Cancel out anything-but-newline followed by backspace
+                    output["text"] = _RGX_BACKSPACE.sub("", output["text"])
+                # Replace all carriage returns not followed by newline
+                output["text"] = _RGX_CARRIAGERETURN.sub("", output["text"])
+
+            # We also want to ensure stdout and stderr are always in the same consecutive order,
+            # because they are asynchronous, so order isn't guaranteed.
+            for i, output in enumerate(new_outputs):
+                if output["output_type"] == "stream" and output["name"] == "stderr":
+                    if (
+                        len(new_outputs) >= i + 2
+                        and new_outputs[i + 1]["output_type"] == "stream"
+                        and new_outputs[i + 1]["name"] == "stdout"
+                    ):
+                        stdout = new_outputs.pop(i + 1)
+                        new_outputs.insert(i, stdout)
+
+            cell.outputs = new_outputs
+
         self.nb['cells'][cell_index] = cell
         return cell
 
@@ -1091,6 +1128,7 @@ class NotebookClient(LoggingConfigurable):
     def output(
         self, outs: t.List, msg: t.Dict, display_id: str, cell_index: int
     ) -> t.Optional[NotebookNode]:
+        """Handle output."""
 
         msg_type = msg['msg_type']
         out: t.Optional[NotebookNode] = None
@@ -1127,7 +1165,7 @@ class NotebookClient(LoggingConfigurable):
         return out
 
     def clear_output(self, outs: t.List, msg: t.Dict, cell_index: int) -> None:
-
+        """Clear output."""
         content = msg['content']
 
         parent_msg_id = msg['parent_header'].get('msg_id')
@@ -1147,13 +1185,13 @@ class NotebookClient(LoggingConfigurable):
             self.clear_display_id_mapping(cell_index)
 
     def clear_display_id_mapping(self, cell_index: int) -> None:
-
+        """Clear a display id mapping for a cell."""
         for _, cell_map in self._display_id_map.items():
             if cell_index in cell_map:
                 cell_map[cell_index] = []
 
     def handle_comm_msg(self, outs: t.List, msg: t.Dict, cell_index: int) -> None:
-
+        """Handle a comm message."""
         content = msg['content']
         data = content['data']
         if self.store_widget_state and 'state' in data:  # ignore custom msg'es
@@ -1225,6 +1263,7 @@ class NotebookClient(LoggingConfigurable):
         assert removed_hook == hook
 
     def on_comm_open_jupyter_widget(self, msg: t.Dict) -> t.Optional[t.Any]:
+        """Handle a jupyter widget comm open."""
         content = msg['content']
         data = content['data']
         state = data['state']
